__all__ = (
    'ConfigurationDict', 'get_default_configuration', 'get_default_config_path',
    'get_default_data_path', 'create_executor', 'resolve_path',
    'normalize_path', 'get_mime_type', 'is_valid_name', 'get_utc_now',
)

from typing import TypedDict, Type
import concurrent.futures
import datetime
import importlib
import mimetypes
import multiprocessing
import pathlib
import signal

from .types import PathOrString


class ConfigurationDict(TypedDict):

    version: int
    driver: str
    database: str
    middleware: list[str]


def get_default_configuration() -> ConfigurationDict:
    return {
        'version': 1,
        'driver': '',
        'database': '',
        'middleware': [],
    }


def get_default_config_path() -> pathlib.Path:
    path = pathlib.Path('~/.config')
    path = path.expanduser()
    path = path / 'wcpan.drive'
    return path


def get_default_data_path() -> pathlib.Path:
    path = pathlib.Path('~/.local/share')
    path = path.expanduser()
    path = path / 'wcpan.drive'
    return path


def create_executor() -> concurrent.futures.Executor:
    if multiprocessing.get_start_method() == 'spawn':
        return concurrent.futures.ProcessPoolExecutor(initializer=initialize_worker)
    else:
        return concurrent.futures.ProcessPoolExecutor()


def initialize_worker() -> None:
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def resolve_path(
    from_: pathlib.PurePath,
    to: pathlib.PurePath,
) -> pathlib.PurePath:
    rv = from_
    for part in to.parts:
        if part == '.':
            continue
        elif part == '..':
            rv = rv.parent
        else:
            rv = rv / part
    return rv


def normalize_path(path: pathlib.PurePath) -> pathlib.PurePath:
    if not path.is_absolute():
        raise ValueError('only accepts absolute path')
    rv = []
    for part in path.parts:
        if part == '.':
            continue
        elif part == '..' and rv[-1] != '/':
            rv.pop()
        else:
            rv.append(part)
    return pathlib.PurePath(*rv)


def import_class(class_path: str) -> Type:
    module_path, class_name = class_path.rsplit('.', 1)
    module = importlib.import_module(module_path)
    class_ = getattr(module, class_name)
    return class_


def get_mime_type(path: PathOrString) -> str:
    type_, dummy_encoding = mimetypes.guess_type(path)
    if not type_:
        return 'application/octet-stream'
    return type_


def is_valid_name(name: str) -> bool:
    if name.find('\\') >= 0:
        return False
    path = pathlib.Path(name)
    return path.name == name


def get_utc_now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)



from typing import Optional, Pattern
import asyncio
import concurrent.futures
import functools
import pathlib
import re
import sqlite3

from .exceptions import CacheError
from .types import Node, ChangeDict


SQL_CREATE_TABLES = [
    '''
    CREATE TABLE metadata (
        key TEXT NOT NULL,
        value TEXT,
        PRIMARY KEY (key)
    );
    ''',

    '''
    CREATE TABLE nodes (
        id TEXT NOT NULL,
        name TEXT,
        trashed BOOLEAN,
        created INTEGER,
        modified INTEGER,
        PRIMARY KEY (id)
    );
    ''',
    'CREATE INDEX ix_nodes_names ON nodes(name);',
    'CREATE INDEX ix_nodes_trashed ON nodes(trashed);',
    'CREATE INDEX ix_nodes_created ON nodes(created);',
    'CREATE INDEX ix_nodes_modified ON nodes(modified);',

    '''
    CREATE TABLE files (
        id TEXT NOT NULL,
        mime_type TEXT,
        hash TEXT,
        size INTEGER,
        PRIMARY KEY (id),
        FOREIGN KEY (id) REFERENCES nodes (id)
    );
    ''',
    'CREATE INDEX ix_files_mime_type ON files(mime_type);',

    '''
    CREATE TABLE parentage (
        parent TEXT NOT NULL,
        child TEXT NOT NULL,
        PRIMARY KEY (parent, child),
        FOREIGN KEY (parent) REFERENCES nodes (id),
        FOREIGN KEY (child) REFERENCES nodes (id)
    );
    ''',
    'CREATE INDEX ix_parentage_parent ON parentage(parent);',
    'CREATE INDEX ix_parentage_child ON parentage(child);',

    '''
    CREATE TABLE images (
        id TEXT NOT NULL,
        width INTEGER NOT NULL,
        height INTEGER NOT NULL,
        PRIMARY KEY (id),
        FOREIGN KEY (id) REFERENCES nodes (id)
    );
    ''',

    '''
    CREATE TABLE videos (
        id TEXT NOT NULL,
        width INTEGER NOT NULL,
        height INTEGER NOT NULL,
        ms_duration INTEGER NOT NULL,
        PRIMARY KEY (id),
        FOREIGN KEY (id) REFERENCES nodes (id)
    );
    ''',

    '''
    CREATE TABLE private (
        id TEXT NOT NULL,
        key TEXT NOT NULL,
        value TEXT
    );
    ''',
    'CREATE INDEX ix_private_id ON private(id);',
    'CREATE INDEX ix_private_key ON private(key);',

    'PRAGMA user_version = 4;',
]


CURRENT_SCHEMA_VERSION = 4


class Cache(object):

    def __init__(self, dsn: str, pool: concurrent.futures.Executor) -> None:
        self._dsn = dsn
        self._pool = pool
        self._raii = None

    async def __aenter__(self) -> 'Cache':
        await self._bg(initialize)
        return self

    async def __aexit__(self, type_, value, traceback) -> bool:
        pass

    async def get_root_id(self) -> str:
        try:
            return await self.get_metadata('root_id')
        except KeyError:
            raise CacheError('invalid cache')

    async def get_root_node(self) -> Node:
        root_id = await self.get_root_id()
        return await self.get_node_by_id(root_id)

    async def get_metadata(self, key: str) -> str:
        return await self._bg(get_metadata, key)

    async def set_metadata(self, key: str, value: str) -> None:
        return await self._bg(set_metadata, key, value)

    async def get_node_by_id(self, node_id: str) -> Optional[Node]:
        return await self._bg(get_node_by_id, node_id)

    async def get_node_by_path(self, path: pathlib.PurePath) -> Optional[Node]:
        try:
            return await self._bg(get_node_by_path, path)
        except KeyError:
            raise CacheError('invalid cache')

    async def get_path_by_id(self, node_id: str) -> Optional[pathlib.PurePath]:
        return await self._bg(get_path_by_id, node_id)

    async def get_node_by_name_from_parent_id(self,
        name: str,
        parent_id: str
    ) -> Optional[Node]:
        return await self._bg(get_node_by_name_from_parent_id, name, parent_id)

    async def get_children_by_id(self, node_id: str) -> list[Node]:
        return await self._bg(get_children_by_id, node_id)

    async def get_trashed_nodes(self) -> list[Node]:
        return await self._bg(get_trashed_nodes)

    async def apply_changes(self,
        changes: list[ChangeDict],
        check_point: str,
    ) -> None:
        return await self._bg(apply_changes, changes, check_point)

    async def insert_node(self, node: Node) -> None:
        return await self._bg(insert_node, node)

    async def find_nodes_by_regex(self, pattern: str) -> list[Node]:
        return await self._bg(find_nodes_by_regex, pattern)

    async def get_uploaded_size(self, begin: int, end: int) -> int:
        return await self._bg(get_uploaded_size, begin, end)

    async def find_orphan_nodes(self) -> list[Node]:
        return await self._bg(find_orphan_nodes)

    async def find_multiple_parents_nodes(self) -> list[Node]:
        return await self._bg(find_multiple_parents_nodes)

    async def _bg(self, fn, *args):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self._pool, fn, self._dsn, *args)


class Database(object):

    def __init__(self, dsn: str) -> None:
        self._dsn = dsn
        self._db = None

    def __enter__(self) -> sqlite3.Connection:
        self._db = sqlite3.connect(self._dsn)
        self._db.row_factory = sqlite3.Row
        # self._db.execute('PRAGMA foreign_keys = 1;')
        return self._db

    def __exit__(self, type_, value, traceback) -> bool:
        self._db.close()
        self._db = None


class ReadOnly(object):

    def __init__(self, db: sqlite3.Connection) -> None:
        self._db = db

    def __enter__(self) -> sqlite3.Cursor:
        self._cursor = self._db.cursor()
        return self._cursor

    def __exit__(self, type_, value, traceback) -> bool:
        self._cursor.close()


class ReadWrite(object):

    def __init__(self, db: sqlite3.Connection) -> None:
        self._db = db

    def __enter__(self) -> sqlite3.Cursor:
        self._cursor = self._db.cursor()
        return self._cursor

    def __exit__(self, type_, value, traceback) -> bool:
        if self._db.in_transaction:
            if type_ is None:
                self._db.commit()
            else:
                self._db.rollback()
        self._cursor.close()


def initialize(dsn: str):
    with Database(dsn) as db:
        try:
            # initialize table
            with ReadWrite(db) as query:
                for sql in SQL_CREATE_TABLES:
                    query.execute(sql)
        except sqlite3.OperationalError as e:
            pass

        # check the schema version
        with ReadOnly(db) as query:
            query.execute('PRAGMA user_version;')
            rv = query.fetchone()
        version = int(rv[0])

        if version < CURRENT_SCHEMA_VERSION:
            raise CacheError((
                'impossible to migrate from old schema prior to '
                f'version {CURRENT_SCHEMA_VERSION}, please rebuild the cache'
            ))


def get_metadata(dsn: str, key: str) -> str:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        return inner_get_metadata(query, key)


def set_metadata(dsn: str, key: str, value: str) -> None:
    with Database(dsn) as db, \
         ReadWrite(db) as query:
        inner_set_metadata(query, key, value)


def get_node_by_id(dsn: str, node_id: str) -> Optional[Node]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        return inner_get_node_by_id(query, node_id)


def get_node_by_path(
    dsn: str,
    path: pathlib.PurePath,
) -> Optional[Node]:
    parts = path.parts[1:]
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        node_id = inner_get_metadata(query, 'root_id')

        for part in parts:
            query.execute('''
                SELECT nodes.id AS id
                FROM parentage
                    INNER JOIN nodes ON parentage.child=nodes.id
                WHERE parentage.parent=? AND nodes.name=?
            ;''', (node_id, part))
            rv = query.fetchone()
            if not rv:
                return None
            node_id = rv['id']

        node = inner_get_node_by_id(query, node_id)
    return node


def get_path_by_id(dsn: str, node_id: str) -> Optional[pathlib.PurePath]:
    parts = []
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        while True:
            query.execute('''
                SELECT name
                FROM nodes
                WHERE id=?
            ;''', (node_id,))
            rv = query.fetchone()
            if not rv:
                raise CacheError(f'cannot find name for {node_id}')

            name = rv['name']
            if not name:
                parts.insert(0, '/')
                break
            parts.insert(0, name)

            query.execute('''
                SELECT parent
                FROM parentage
                WHERE child=?
            ;''', (node_id,))
            rv = query.fetchone()
            if not rv:
                # orphan node
                break
            node_id = rv['parent']

    path = pathlib.PurePath(*parts)
    return path


def get_node_by_name_from_parent_id(
    dsn: str,
    name: str,
    parent_id: str
) -> Optional[Node]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT nodes.id AS id
            FROM nodes
                INNER JOIN parentage ON parentage.child=nodes.id
            WHERE parentage.parent=? AND nodes.name=?
        ;''', (parent_id, name))
        rv = query.fetchone()

        if not rv:
            return None

        node = inner_get_node_by_id(query, rv['id'])
    return node


def get_children_by_id(dsn: str, node_id: str) -> list[Node]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT child
            FROM parentage
            WHERE parent=?
        ;''', (node_id,))
        rv = query.fetchall()

        children = [inner_get_node_by_id(query, _['child']) for _ in rv]
    return children


def get_trashed_nodes(dsn: str) -> list[Node]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT id
            FROM nodes
            WHERE trashed=?
        ;''', (True,))
        rv = query.fetchall()

        children = [inner_get_node_by_id(query, _['id']) for _ in rv]
    return children


def apply_changes(
    dsn: str,
    changes: list[ChangeDict],
    check_point: str,
) -> None:
    with Database(dsn) as db, \
         ReadWrite(db) as query:
        for change in changes:
            is_removed = change['removed']
            if is_removed:
                inner_delete_node_by_id(query, change['id'])
                continue

            node = Node.from_dict(change['node'])
            inner_insert_node(query, node)

        inner_set_metadata(query, 'check_point', check_point)


def insert_node(dsn: str, node: Node) -> None:
    with Database(dsn) as db, \
         ReadWrite(db) as query:
        inner_insert_node(query, node)

        if not node.name:
            inner_set_metadata(query, 'root_id', node.id_)


def find_nodes_by_regex(dsn: str, pattern: str) -> list[Node]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        fn = functools.partial(sqlite3_regexp, re.compile(pattern, re.I))
        db.create_function('REGEXP', 2, fn)
        query.execute('SELECT id FROM nodes WHERE name REGEXP ?;', (pattern,))
        rv = query.fetchall()
        rv = [inner_get_node_by_id(query, _['id']) for _ in rv]
    return rv


def get_uploaded_size(dsn: str, begin: int, end: int) -> int:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT SUM(size) AS sum
            FROM files
                INNER JOIN nodes ON files.id = nodes.id
            where created >= ? AND created < ?
        ;''', (begin, end))
        rv = query.fetchone()
        if not rv:
            return 0
        if rv['sum'] is None:
            return 0
        return rv['sum']


def find_orphan_nodes(dsn: str) -> list[Node]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT nodes.id AS id
            FROM parentage
                LEFT OUTER JOIN nodes ON parentage.child=nodes.id
            WHERE parentage.parent IS NULL
        ;''')
        rv = query.fetchall()
        rv = [inner_get_node_by_id(query, _['id']) for _ in rv]
    return rv


def find_multiple_parents_nodes(dsn: str) -> list[Node]:
    with Database(dsn) as db, \
         ReadOnly(db) as query:
        query.execute('''
            SELECT child, COUNT(child) AS parent_count
            FROM parentage
            GROUP BY child
            HAVING parent_count > 1
        ;''')
        rv = query.fetchall()
        rv = [inner_get_node_by_id(query, _['child']) for _ in rv]
    return rv


def inner_get_metadata(query: sqlite3.Cursor, key: str) -> str:
    query.execute('SELECT value FROM metadata WHERE key = ?;', (key,))
    rv = query.fetchone()
    if not rv:
        raise KeyError(key)
    return rv['value']


def inner_set_metadata(query: sqlite3.Cursor, key: str, value: str) -> None:
    query.execute('''
        INSERT OR REPLACE INTO metadata
        VALUES (?, ?)
    ;''', (key, value))


def inner_get_node_by_id(
    query: sqlite3.Cursor,
    node_id: str,
) -> Optional[Node]:
    query.execute('''
        SELECT name, trashed, created, modified
        FROM nodes
        WHERE id=?
    ;''', (node_id,))
    rv = query.fetchone()
    if not rv:
        return None
    node = dict(rv)
    node['id'] = node_id

    query.execute('''
        SELECT mime_type, hash, size
        FROM files
        WHERE id=?
    ;''', (node_id,))
    rv = query.fetchone()
    is_folder = rv is None
    node['is_folder'] = is_folder
    node['mime_type'] = None if is_folder else rv['mime_type']
    node['hash'] = None if is_folder else rv['hash']
    node['size'] = None if is_folder else rv['size']

    query.execute('''
        SELECT parent
        FROM parentage
        WHERE child=?
    ;''', (node_id,))
    rv = query.fetchall()
    node['parent_list'] = [_['parent'] for _ in rv]

    query.execute('''
        SELECT width, height
        FROM images
        WHERE id=?
    ;''', (node_id,))
    rv = query.fetchone()
    node['image'] = {
        'width': rv['width'],
        'height': rv['height'],
    } if rv else None

    query.execute('''
        SELECT width, height, ms_duration
        FROM videos
        WHERE id=?
    ;''', (node_id,))
    rv = query.fetchone()
    node['video'] = {
        'width': rv['width'],
        'height': rv['height'],
        'ms_duration': rv['ms_duration'],
    } if rv else None

    query.execute('''
        SELECT key, value
        FROM private
        WHERE id=?;
    ''', (node_id,))
    rv = query.fetchall()
    node['private'] = None if not rv else {_['key']: _['value'] for _ in rv}

    node = Node.from_dict(node)
    return node


def inner_insert_node(query: sqlite3.Cursor, node: Node) -> None:
    # add this node
    query.execute('''
        INSERT OR REPLACE INTO nodes
        (id, name, trashed, created, modified)
        VALUES
        (?, ?, ?, ?, ?)
    ;''', (node.id_, node.name, node.trashed,
           node.created.int_timestamp, node.modified.int_timestamp))

    # add file information
    if not node.is_folder:
        query.execute('''
            INSERT OR REPLACE INTO files
            (id, mime_type, hash, size)
            VALUES
            (?, ?, ?, ?)
        ;''', (node.id_, node.mime_type, node.hash_, node.size))

    # remove old parentage
    query.execute('''
        DELETE FROM parentage
        WHERE child=?
    ;''', (node.id_,))
    # add parentage if there is any
    if node.parent_list:
        for parent in node.parent_list:
            query.execute('''
                INSERT INTO parentage
                (parent, child)
                VALUES
                (?, ?)
            ;''', (parent, node.id_))

    # add image information
    if node.is_image:
        query.execute('''
            INSERT OR REPLACE INTO images
            (id, width, height)
            VALUES
            (?, ?, ?)
        ;''', (node.id_, node.image_width, node.image_height))

    # add video information
    if node.is_video:
        query.execute('''
            INSERT OR REPLACE INTO videos
            (id, width, height, ms_duration)
            VALUES
            (?, ?, ?, ?)
        ;''', (node.id_, node.video_width, node.video_height,
               node.video_ms_duration))

    # remove old private
    query.execute('''
        DELETE FROM private
        WHERE id=?
    ;''', (node.id_,))
    # add private information if any
    if node.private:
        for key, value in node.private.items():
            query.execute('''
                INSERT INTO private
                (id, key, value)
                VALUES
                (?, ?, ?)
            ;''', (node.id_, key, value))


def inner_delete_node_by_id(query: sqlite3.Cursor, node_id: str) -> None:
    # remove from private
    query.execute('''
        DELETE FROM private
        WHERE id=?
    ;''', (node_id,))

    # remove from videos
    query.execute('''
        DELETE FROM videos
        WHERE id=?
    ;''', (node_id,))

    # remove from images
    query.execute('''
        DELETE FROM images
        WHERE id=?
    ;''', (node_id,))

    # disconnect parents
    query.execute('''
        DELETE FROM parentage
        WHERE child=? OR parent=?
    ;''', (node_id, node_id))

    # remove from files
    query.execute('''
        DELETE FROM files
        WHERE id=?
    ;''', (node_id,))

    # remove from nodes
    query.execute('''
        DELETE FROM nodes
        WHERE id=?
    ;''', (node_id,))


def sqlite3_regexp(
    pattern: Pattern,
    _: str,
    cell: Optional[str],
) -> bool:
    if cell is None:
        # root node
        return False
    return pattern.search(cell) is not None



import contextlib
import hashlib
import os
import pathlib
from tempfile import TemporaryDirectory
from typing import AsyncIterator, Optional
from unittest.mock import AsyncMock

from .abc import RemoteDriver, ReadableFile, WritableFile, Hasher
from .drive import DriveFactory, ReadOnlyContext
from .types import Node, ChangeDict, PrivateDict, MediaInfo, NodeDict
from .util import get_utc_now


@contextlib.contextmanager
def test_factory(driver_class: str = None, middleware_list: list[str] = None):
    with TemporaryDirectory() as work_folder:
        work_path = pathlib.Path(work_folder)
        config_path = work_path / 'config'
        config_path.mkdir()
        data_path = work_path / 'data'
        data_path.mkdir()

        factory = DriveFactory()
        factory.config_path = config_path
        factory.data_path = data_path
        factory.database = data_path / 'nodes.sqlite'
        factory.driver = 'wcpan.drive.core.test.TestDriver' if driver_class is None else driver_class
        factory.middleware_list = [] if middleware_list is None else middleware_list

        yield factory


class TestDriver(RemoteDriver):

    @classmethod
    def get_version_range(cls):
        return (3, 3)

    def __init__(self, context: ReadOnlyContext) -> None:
        self.mock = MockManager()
        self.pseudo = PseudoManager()
        self._context = context
        self._root: pathlib.Path = None
        self._raii = None

    async def __aenter__(self) -> RemoteDriver:
        with contextlib.ExitStack() as stack:
            root = stack.enter_context(TemporaryDirectory())
            self._root = pathlib.Path(root)
            self._raii = stack.pop_all()
        return self

    async def __aexit__(self, et, ev, tb) -> bool:
        self._root = None
        self._raii.close()
        self._raii = None

    @property
    def remote(self):
        return None

    async def get_initial_check_point(self) -> str:
        return str(self.pseudo.INITIAL_CHECK_POINT)

    async def fetch_root_node(self) -> Node:
        node = self.pseudo.build_node().node
        return node

    async def fetch_changes(self, check_point: str):
        yield str(self.pseudo.check_point), self.pseudo.changes
        self.pseudo.consume()

    async def create_folder(self,
        parent_node: Node,
        folder_name: str,
        *,
        exist_ok: bool,
        private: Optional[PrivateDict],
    ) -> Node:
        await self.mock.create_folder(parent_node, folder_name, private, exist_ok)

        builder = self.pseudo.build_node()
        builder.to_folder(folder_name, parent_node)
        folder = builder.commit()

        return folder

    async def rename_node(self,
        node: Node,
        *,
        new_parent: Optional[Node],
        new_name: Optional[str],
    ) -> Node:
        await self.mock.rename_node(node, new_parent, new_name)

        dict_ = node.to_dict()
        if new_parent:
            dict_['parent_list'] = [new_parent.id_]
        if new_name:
            dict_['name'] = new_name

        self.pseudo.update(dict_)

        return Node.from_dict(dict_)

    async def trash_node(self, node: Node) -> None:
        await self.mock.trash_node(node)

        dict_ = node.to_dict()
        dict_['trashed'] = True

        self.pseudo.update(dict_)

    async def download(self, node: Node) -> ReadableFile:
        await self.mock.download(node)
        return NodeReader(self._root, node)

    async def upload(self,
        parent_node: Node,
        file_name: str,
        *,
        file_size: Optional[int],
        mime_type: Optional[str],
        media_info: Optional[MediaInfo],
        private: Optional[PrivateDict],
    ) -> WritableFile:
        await self.mock.upload(
            parent_node,
            file_name,
            file_size,
            mime_type,
            media_info,
            private,
        )
        return NodeWriter(
            root=self._root,
            pseudo=self.pseudo,
            parent_node=parent_node,
            file_name=file_name,
            mime_type=mime_type,
            media_info=media_info,
        )

    async def get_hasher(self) -> Hasher:
        await self.mock.get_hasher()
        return NodeHasher()

    async def is_authorized(self) -> bool:
        return True

    async def get_oauth_url(self) -> str:
        return ''

    async def set_oauth_token(self, token: str) -> None:
        pass


class MockManager(object):

    def __init__(self):
        self.create_folder = AsyncMock()
        self.rename_node = AsyncMock()
        self.trash_node = AsyncMock()
        self.download = AsyncMock()
        self.upload = AsyncMock()
        self.get_hasher = AsyncMock()

    def reset_all_mocks(self, *, return_value: bool = False, side_effect: bool = False):
        self.create_folder.reset_mock(return_value=return_value, side_effect=side_effect)
        self.rename_node.reset_mock(return_value=return_value, side_effect=side_effect)
        self.trash_node.reset_mock(return_value=return_value, side_effect=side_effect)
        self.download.reset_mock(return_value=return_value, side_effect=side_effect)
        self.upload.reset_mock(return_value=return_value, side_effect=side_effect)
        self.get_hasher.reset_mock(return_value=return_value, side_effect=side_effect)


class PseudoManager(object):

    INITIAL_CHECK_POINT = 1

    def __init__(self):
        self.check_point = 1
        self.changes: list[ChangeDict] = []
        self._id = 0

    def next_id(self) -> str:
        id_ = self._id
        self._id += 1
        return f'__ID_{id_}__'

    def build_node(self) -> 'NodeBuilder':
        return NodeBuilder(self)

    def update(self, dict_: NodeDict) -> None:
        self.changes.append({
            'removed': False,
            'node': dict_,
        })
        self.check_point += 1

    def delete(self, id_: str) -> None:
        self.changes.append({
            'removed': True,
            'id': id_,
        })
        self.check_point += 1

    def consume(self) -> None:
        self.changes = []


class NodeBuilder(object):

    def __init__(self, pseudo: PseudoManager):
        self.pseudo = pseudo
        self.dict: NodeDict = {
            'id': '__ID_ROOT__',
            'name': None,
            'is_folder': True,
            'trashed': False,
            'created': get_utc_now().isoformat(),
            'modified': get_utc_now().isoformat(),
            'parent_list': [],
            'size': None,
            'mime_type': None,
            'hash': None,
            'image': None,
            'video': None,
            'private': None,
        }

    @property
    def node(self):
        return Node.from_dict(self.dict)

    def to_trashed(self, trashed: bool = True):
        self.dict['trashed'] = trashed
        return self

    def to_folder(self, name: str, parent: Node):
        self.dict['id'] = self.pseudo.next_id()
        self.dict['name'] = name
        self.dict['parent_list'] = [parent.id_]
        return self

    def to_file(self, size: int, hash_: str, mime_type: str):
        self.dict['is_folder'] = False
        self.dict['size'] = size
        self.dict['hash'] = hash_
        self.dict['mime_type'] = mime_type
        return self

    def to_image(self, width: int, height: int):
        self.dict['image'] = {
            'width': width,
            'height': height,
        }
        return self

    def to_video(self, width: int, height: int, ms_duration: int):
        self.dict['video'] = {
            'width': width,
            'height': height,
            'ms_duration': ms_duration,
        }
        return self

    def commit(self):
        self.pseudo.update(self.dict)
        return self.node


class NodeHasher(Hasher):

    def __init__(self):
        self._hasher = hashlib.new('md5')

    def update(self, data: bytes) -> None:
        self._hasher.update(data)

    def digest(self) -> bytes:
        return self._hasher.digest()

    def hexdigest(self) -> str:
        return self._hasher.hexdigest()

    def copy(self) -> Hasher:
        return NodeHasher()


class NodeReader(ReadableFile):

    def __init__(self, root: pathlib.Path, node: Node):
        self._root = root
        self._node = node
        self._fin = None
        self._raii = None

    async def __aenter__(self) -> ReadableFile:
        with contextlib.ExitStack() as stack:
            path = self._root / self._node.id_
            self._fin = stack.enter_context(path.open('rb'))
            self._raii = stack.pop_all()
        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        self._fin = None
        self._raii.close()
        self._raii = None

    async def __aiter__(self) -> AsyncIterator[bytes]:
        while True:
            chunk = self._fin.read(8192)
            if not chunk:
                return
            yield chunk

    async def read(self, length: int) -> bytes:
        return self._fin.read(length)

    async def seek(self, offset: int) -> None:
        self._fin.flush()
        self._fin.seek(offset, os.SEEK_SET)

    async def node(self) -> Node:
        return self._node


class NodeWriter(WritableFile):

    def __init__(self,
        root: pathlib.Path,
        pseudo: PseudoManager,
        parent_node: Node,
        file_name: str,
        mime_type: Optional[str],
        media_info: Optional[MediaInfo],
    ):
        self._root = root
        self._pseudo = pseudo
        self._parent_node = parent_node
        self._file_name = file_name
        self._mime_type = mime_type
        self._media_info = media_info
        self._node = None
        self._fout = None
        self._raii = None

    async def __aenter__(self) -> WritableFile:
        self._node = None
        with contextlib.ExitStack() as stack:
            path = self._root / '_'
            self._fout = stack.enter_context(path.open('wb'))
            self._raii = stack.pop_all()
        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        self._fout.flush()
        self._fout = None
        self._raii.close()
        self._raii = None

        path = self._root / '_'
        stat = path.stat()
        hasher = NodeHasher()
        with path.open('rb') as fin:
            while True:
                chunk = fin.read(8092)
                if not chunk:
                    break
                hasher.update(chunk)
        hash_ = hasher.hexdigest()

        builder = self._pseudo.build_node()
        builder.to_folder(self._file_name, self._parent_node)
        builder.to_file(stat.st_size, hash_, self._mime_type)
        if self._media_info:
            if self._media_info.is_image:
                builder.to_image(self._media_info.width, self._media_info.height)
            elif self._media_info.is_video:
                builder.to_video(self._media_info.width, self._media_info.height, self._media_info.ms_duration)
        path.rename(self._root / builder.node.id_)
        self._node = builder.commit()

    async def tell(self) -> int:
        return self._fout.tell()

    async def seek(self, offset: int) -> None:
        self._fout.flush()
        self._fout.seek(offset, os.SEEK_SET)

    async def write(self, chunk: bytes) -> int:
        return self._fout.write(chunk)

    async def node(self) -> Optional[Node]:
        return self._node



from .types import Node


class DriveError(Exception):
    pass


class InvalidRemoteDriverError(DriveError):
    pass


class InvalidMiddlewareError(DriveError):
    pass


# NOTE *MUST* be picklable
class CacheError(DriveError):
    pass


class InvalidNameError(DriveError):

    def __init__(self, name: str) -> None:
        self._name = name

    def __str__(self) -> str:
        return f'invalid name: {self._name}'


class NodeTreeError(DriveError):
    pass


class NodeConflictedError(NodeTreeError):

    def __init__(self, node: Node) -> None:
        self._node = node

    def __str__(self) -> str:
        return f'node already exists: {self.node.name}'

    @property
    def node(self) -> Node:
        return self._node


class ParentNotFoundError(NodeTreeError):

    def __init__(self, id_: str) -> None:
        self._id = id_

    def __str__(self) -> str:
        return f'parent id not found: {self._id}'


class NodeNotFoundError(NodeTreeError):

    def __init__(self, path_or_id: str) -> None:
        self._path_or_id = path_or_id

    def __str__(self) -> str:
        return f'node not found: {self._path_or_id}'


class RootNodeError(NodeTreeError):
    pass


class TrashedNodeError(NodeTreeError):
    pass


class ParentIsNotFolderError(NodeTreeError):
    pass


class LineageError(NodeTreeError):
    pass


class UnauthorizedError(DriveError):
    pass


class DownloadError(DriveError):
    pass


class UploadError(DriveError):
    pass




from __future__ import annotations


__all__ = (
    'ReadableFile', 'Hasher', 'WritableFile', 'Middleware', 'RemoteDriver',
)


from abc import ABCMeta, abstractmethod
from typing import (
    AsyncGenerator,
    AsyncIterator,
    Optional,
    Type,
)
from types import TracebackType

from .types import (
    ChangeDict,
    MediaInfo,
    Node,
    PrivateDict,
)


class ReadableFile(metaclass=ABCMeta):
    '''
    An async readable file interface.

    Should support async iterator and async context manager.

    Can be used like this:
    ```
    async with ReadableFile(...) as fin:
        async for chunk in fin:
            ...
    ```
    '''

    @abstractmethod
    def __aiter__(self) -> AsyncIterator[bytes]:
        ...

    @abstractmethod
    async def __aenter__(self) -> ReadableFile:
        ...

    @abstractmethod
    async def __aexit__(self,
        et: Optional[Type[BaseException]],
        ev: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:
        ...

    @abstractmethod
    async def read(self, length: int) -> bytes:
        '''
        Read at most `length` bytes.
        '''

    @abstractmethod
    async def seek(self, offset: int) -> None:
        '''
        Seek to `offset` position. Always starts from the begining.
        '''

    @abstractmethod
    async def node(self) -> Node:
        '''
        Get the node being read.
        '''


class Hasher(metaclass=ABCMeta):
    '''
    Hash calculator.

    MUST be pickleable to work with multi-processes.
    '''

    @abstractmethod
    def update(self, data: bytes) -> None:
        '''
        Put `data` into the stream.
        '''

    @abstractmethod
    def digest(self) -> bytes:
        '''
        Get raw digest.
        '''

    @abstractmethod
    def hexdigest(self) -> str:
        '''
        Get hex digest.
        '''

    @abstractmethod
    def copy(self) -> Hasher:
        '''
        Return a copy to self. Does not require clone the state.
        '''


class WritableFile(metaclass=ABCMeta):
    '''
    An async writable file interface.

    Should support and async context manager.

    Can be used like this:
    ```
    async with WritableFile(...) as fout:
        ...
    ```
    '''

    @abstractmethod
    async def __aenter__(self) -> WritableFile:
        ...

    @abstractmethod
    async def __aexit__(self,
        et: Optional[Type[BaseException]],
        ev: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:
        ...

    @abstractmethod
    async def tell(self) -> int:
        '''
        Get current position.
        '''

    @abstractmethod
    async def seek(self, offset: int) -> None:
        '''
        Seek to `offset` position. Always starts from the begining.
        '''

    @abstractmethod
    async def write(self, chunk: bytes) -> int:
        '''
        Writes `chunk` to the stream.
        Returns actual written byte length.
        '''

    @abstractmethod
    async def node(self) -> Optional[Node]:
        '''
        Get the wrote node. May be `None` if write failed.
        '''


class RemoteDriver(metaclass=ABCMeta):
    '''
    Provides actions to cloud drives.

    Must be used with async context manager.
    '''

    @classmethod
    @abstractmethod
    def get_version_range(cls) -> tuple[int, int]:
        '''
        Get competible API version range for this driver.

        The tuple is (minimal, maximum), inclusive.
        '''

    @property
    @abstractmethod
    def remote(self) -> Optional[RemoteDriver]:
        '''
        Get the decorated remote driver, if any.
        '''

    @abstractmethod
    async def __aenter__(self) -> RemoteDriver:
        ...

    @abstractmethod
    async def __aexit__(self,
        et: Optional[Type[BaseException]],
        ev: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:
        ...

    @abstractmethod
    async def get_initial_check_point(self) -> str:
        '''
        Get the initial check point.
        '''

    @abstractmethod
    async def fetch_root_node(self) -> Node:
        '''
        Fetch the root node.
        '''

    @abstractmethod
    async def fetch_changes(self,
        check_point: str,
    ) -> AsyncGenerator[tuple[str, list[ChangeDict]], None]:
        '''
        Fetch changes starts from `check_point`.

        Will be used like this:
        ```
        async for next_check_point, changes in self.fetch_changes('...'):
            ...
        ```
        So you should yield a page for every iteration.
        '''

    @abstractmethod
    async def create_folder(self,
        parent_node: Node,
        folder_name: str,
        *,
        exist_ok: bool,
        private: Optional[PrivateDict],
    ) -> Node:
        '''
        Create a folder.

        `parent_node` should be a folder you want to put this folder in.

        `folder_name` will be the name of the folder.

        `private` is an optional metadata, you can decide how to place this for
        each services.

        If `exist_ok` is `False`, you should not create the folder if it is
        already exists, and raise an exception.

        Will return the created node.
        '''

    @abstractmethod
    async def rename_node(self,
        node: Node,
        *,
        new_parent: Optional[Node],
        new_name: Optional[str],
    ) -> Node:
        '''
        Rename a node, or move to another folder, or do both.

        `node` is the node to be modified.

        `new_parent` is the new parent folder. `None` means don't move the node.

        `new_name` is the new node name. `None` means don't rename the node.
        '''

    @abstractmethod
    async def trash_node(self, node: Node) -> None:
        '''
        Trash the node.

        Should raise exception if failed.
        '''

    @abstractmethod
    async def download(self, node: Node) -> ReadableFile:
        '''
        Download the node.

        Will return a `ReadableFile` which is a file-like object.
        '''

    @abstractmethod
    async def upload(self,
        parent_node: Node,
        file_name: str,
        *,
        file_size: Optional[int],
        mime_type: Optional[str],
        media_info: Optional[MediaInfo],
        private: Optional[PrivateDict],
    ) -> WritableFile:
        '''
        Upload a file.

        `parent_node` is the target folder.

        `file_name` is required.

        `file_size` can be `None`, for cases that the file size is unavailable.
        e.g. The uploading file is from a stream.

        `mime_type`, `media_info` and `private` are optional. It is your choice
        to decide how to place these properties.
        '''

    @abstractmethod
    async def get_hasher(self) -> Hasher:
        '''
        Get a hash calculator.
        '''

    @abstractmethod
    async def is_authorized(self) -> bool:
        '''
        Is OAuth 2.0 authorized.
        '''

    @abstractmethod
    async def get_oauth_url(self) -> str:
        '''
        Get OAuth 2.0 URL.
        '''

    @abstractmethod
    async def set_oauth_token(self, token: str) -> None:
        '''
        Set OAuth 2.0 token.
        '''


Middleware = RemoteDriver
